@@ -11,7 +11,6 @@ use File::Spec;
 use File::Path;
 use File::Copy;
 use File::Basename qw(dirname basename);
-use Struct::Compare;
 use Crypt::PasswdMD5 qw(unix_md5_crypt);
 require Exporter;
 #======================================================================
@@ -38,6 +37,9 @@ use constant UMASK		=> 0022;
 use constant PERM_PWD	=> 0644;
 use constant PERM_GRP	=> 0644;
 use constant PERM_SHD	=> 0400;
+use constant PERM_SHD_W	=> 0600;
+use constant PERM_GSD	=> 0440;
+use constant PERM_GSD_W => 0640;
 use constant PATH		=>  qr/^[\w\+_\040\#\(\)\{\}\[\]\/\-\^,\.:;&%@\\~]+\$?$/;
 #======================================================================
 my $_CHECK = {
@@ -46,7 +48,7 @@ my $_CHECK = {
 	'uid'		=> sub { return if not defined $_[0] or $_[0] !~ /^[0-9]+$/o; TRUE },
 	'home'		=> sub { return if not defined $_[0] or $_[0] !~ PATH; TRUE },
 	'shell'		=> sub { return if not defined $_[0] or $_[0] !~ PATH; TRUE },
-	'gecos'		=> sub { return if not defined $_[0] or $_[0] !~ /^[^:]+$/o; TRUE },
+	'gecos'		=> sub { return if not defined $_[0] or $_[0] !~ /^[^:]*$/o; TRUE },
 	'passwd' 	=> sub { return if not defined $_[0]; TRUE},
 };
 #======================================================================
@@ -73,6 +75,16 @@ sub new {
 	return $self;
 }
 #======================================================================
+sub array_compare {
+    my ($aname, $a, $bname, $b) = @_;
+    my @aonly = grep {!($_ ~~ @$b)} @$a;
+    my @bonly = grep {!($_ ~~ @$a)} @$b;
+    join("; ", grep {length} (
+        (@aonly ? "only in $aname: ".join(", ", @aonly) : ""),
+        (@bonly ? "only in $bname: ".join(", ", @bonly) : ""),
+    ));
+}
+#======================================================================
 sub check_sanity {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my $quiet = shift;
@@ -86,7 +98,7 @@ sub check_sanity {
 		carp(q/Insecure permissions to group file!/)	and sleep(0) if ((stat($self->group_file)  )[2] & 07777) != PERM_GRP;
 		carp(q/Insecure permissions to passwd file!/)	and sleep(0) if ((stat($self->passwd_file) )[2] & 07777) != PERM_PWD;
 		carp(q/Insecure permissions to shadow file!/)	and sleep(0) if ((stat($self->shadow_file) )[2] & 07777) != PERM_SHD;
-		carp(q/Insecure permissions to gshadow file!/)	and sleep(0) if ((stat($self->gshadow_file))[2] & 07777) != PERM_GRP;
+		carp(q/Insecure permissions to gshadow file!/)	and sleep(0) if ((stat($self->gshadow_file))[2] & 07777) != PERM_GSD;
 	}
 
 	my %filenames = ( shadow => $self->shadow_file, passwd => $self->passwd_file, group => $self->group_file, gshadow => $self->gshadow_file );
@@ -97,8 +109,9 @@ sub check_sanity {
 		}
 	}
 
-	unless(compare([$self->users()], [$self->users_from_shadow()])){
-		carp(qq/\nYour ENVIRONMENT IS INSANE! Users in files "/.$self->passwd_file().q/" and "/.$self->shadow_file().qq/ are diffrent!!!\nI'll continue, but it is YOUR RISK! You'll probably go into BIG troubles!\n\n/);
+	my $diff = array_compare("passwd", [$self->users()], "shadow", [$self->users_from_shadow()]);
+	if ($diff) {
+		carp(qq/\nYour ENVIRONMENT IS INSANE! Users in files "/.$self->passwd_file().q/" and "/.$self->shadow_file().qq/" are different!!!\n\n $diff\n\nI'll continue, but it is YOUR RISK! You'll probably go into BIG troubles!\n\n/);
 		warn "\a\n";
 		sleep 0;
 	}
@@ -247,8 +260,8 @@ sub del {
 
 	# remove from passwd
 	my $tmp = $self->passwd_file.'.tmp';
-	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
-	open(my $ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $!"; return };
+	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open(my $ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp (1) : $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_PWD, $ch;
 	while(my $line = <$fh>){
 		my ($user, undef, undef, $gid) = split(/:/,$line, 5);
@@ -261,26 +274,26 @@ sub del {
 		}
 	}
 	close($fh);close($ch);
-	move($tmp, $self->passwd_file()) or do { $errstr = "Can't replace passwd file ".$self->passwd_file.": $!"; return };
+	move($tmp, $self->passwd_file()) or do { $errstr = "Can't replace passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	# remove from shadow
 	$tmp = $self->shadow_file.'.tmp';
-	open($fh, '<', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $!"; return };
-	open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $!"; return };
+	open($fh, '<', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_SHD, $ch;
 	while(my $line = <$fh>){
 		next if (split(/:/,$line,2))[0] =~ $regexp;
 		print $ch $line;
 	}
 	close($fh);close($ch);
-	move($tmp, $self->shadow_file()) or do { $errstr = "Can't replace shadow file ".$self->shadow_file.": $!"; return };
+	move($tmp, $self->shadow_file()) or do { $errstr = "Can't replace shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	# remove from group
 	my $gids = '^'.join('$|^',@gids).'$';
 	$gids = qr/$gids/;
 	$tmp = $self->group_file.'.tmp';
-	open($fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
-	open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $!"; return };
+	open($fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_GRP, $ch;
 	while(my $line = <$fh>){
 		chomp $line;
@@ -289,14 +302,14 @@ sub del {
 		print $ch join(q/:/, $name, $passwd, $gid, $users),"\n";
 	}
 	close($fh);close($ch);
-	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $!"; return };
+	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	# remove from gshadow
 	if(-f $self->gshadow_file){
 		$tmp = $self->gshadow_file.'.tmp';
-		open($fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $!"; return };
-		open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $!"; return };
-		chmod PERM_SHD, $ch;
+		open($fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+		open($ch, '>', $tmp) or do { $errstr = "Can't open temp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
+		chmod PERM_GSD, $ch;
 		while(my $line = <$fh>){
 			chomp $line;
 			my ($name, $passwd, $gid, $users) = split(/:/,$line,4);
@@ -304,7 +317,7 @@ sub del {
 			print $ch join(q/:/, $name, $passwd, $gid, $users),"\n";
 		}
 		close($fh);close($ch);
-		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $!"; return };
+		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	}
 
 	umask $umask;
@@ -334,12 +347,12 @@ sub _set {
 	my $mode	=	$file eq $self->passwd_file()	?	PERM_PWD	:
 					$file eq $self->group_file()	?	PERM_GRP	:
 					$file eq $self->shadow_file()	?	PERM_SHD	:
-														PERM_SHD	;
+														PERM_GSD	;
 
 	$count ||= 6;
 	my $tmp = $file.'.tmp';
-	open(my $fh, '<', $file) or do { $errstr = "Can't open file $file: $!"; return };
-	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+	open(my $fh, '<', $file) or do { $errstr = "Can't open file $file: $! (".__FILE__." line ".__LINE__.")"; return };
+	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod $mode, $ch;
 	my $ret;
 	while(<$fh>){
@@ -355,7 +368,7 @@ sub _set {
 		}
 	}
 	close($fh);close($ch);
-	move($tmp, $file) or do { $errstr = "Can't replace file $file: $!"; return };
+	move($tmp, $file) or do { $errstr = "Can't replace file $file: $! (".__FILE__." line ".__LINE__.")"; return };
 
 	umask $umask;
 
@@ -374,7 +387,7 @@ sub _get {
                 return;
 	}
 
-	open(my $fh, '<', $file) or do { $errstr = "can't open file $file: $!"; return };
+	open(my $fh, '<', $file) or do { $errstr = "can't open file $file: $! (".__FILE__." line ".__LINE__.")"; return };
 	while(<$fh>){
 		my @a = split /:/;
 		next if $a[0] ne $user;
@@ -481,8 +494,8 @@ sub rename {
 	my $umask = umask $self->{'umask'};
 
 	my $tmp = $self->group_file.'.tmp';
-	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
-	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_GRP, $ch;
 	while(my $line = <$fh>){
 		chomp $line;
@@ -491,12 +504,12 @@ sub rename {
 		print $ch join(q/:/, $name, $passwd, $gid, $users),"\n";
 	}
 	close($fh);close($ch);
-	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $!"; return };
+	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	if(-f $self->gshadow_file){
 		my $tmp = $self->gshadow_file.'.tmp';
-		open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $!"; return };
-		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+		open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 		chmod PERM_PWD, $ch;
 		while(my $line = <$fh>){
 			chomp $line;
@@ -505,7 +518,7 @@ sub rename {
 			print $ch join(q/:/, $name, $passwd, $gid, $users),"\n";
 		}
 		close($fh);close($ch);
-		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $!"; return };
+		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 ;
 	}
 
@@ -519,7 +532,7 @@ sub rename {
 sub maxgid {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my $max = 0;
-	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
+	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	while(<$fh>){
 		my $tmp = (split(/:/,$_))[3];
 		$max = $tmp > $max ? $tmp : $max;
@@ -532,7 +545,7 @@ sub maxgid {
 sub maxuid {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my $max = 0;
-	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
+	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	while(<$fh>){
 		my $tmp = (split(/:/,$_))[2];
 		$max = $tmp > $max ? $tmp : $max;
@@ -547,7 +560,7 @@ sub _exists {
 	return if scalar @_ != 3;
 	my ($file, $pos, $val) = @_;
 
-	open(my $fh, '<', $file) or do { $errstr = "Can't open file $file: $!"; return };
+	open(my $fh, '<', $file) or do { $errstr = "Can't open file $file: $! (".__FILE__." line ".__LINE__.")"; return };
 	$errstr = "";
         while(<$fh>){
 		my @a = split /:/;
@@ -590,7 +603,7 @@ sub user {
 	}
 
 	if(scalar @_ != 7){
-		open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
+		open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 		while(<$fh>){
 			my @a = split /:/;
 			next if $a[0] ne $user[0];
@@ -624,8 +637,8 @@ sub user {
 
 	my $mod;
 	my $tmp = $self->passwd_file.'.tmp';
-	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
-	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_PWD, $ch;
 	while(<$fh>){
 		my @a = split /:/;
@@ -637,12 +650,13 @@ sub user {
 	close($fh);
 	print $ch join(q/:/, @user),"\n" unless $mod;
 	close($ch);
-	move($tmp, $self->passwd_file()) or do { $errstr = "Can't replace passwd file ".$self->passwd_file.": $!"; return };
+	move($tmp, $self->passwd_file()) or do { $errstr = "Can't replace passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	# user already exists
 	if($mod){ $self->passwd($user[0], $passwd); }
 	else{
-		open(my $fh, '>>', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $!"; return };
+		chmod PERM_SHD_W, $self->shadow_file() or do { $errstr = "Can't chmod shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+		open(my $fh, '>>', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 		chmod PERM_SHD, $fh;
 		print $fh join(q/:/, $user[0], $passwd, int(time()/DAY), ('') x 5, "\n");
 		close($fh);
@@ -656,7 +670,7 @@ sub user {
 sub users {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my @a;
-	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $!"; return };
+	open(my $fh, '<', $self->passwd_file()) or do { $errstr = "Can't open passwd file ".$self->passwd_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	push @a, (split(/:/,$_))[0] while <$fh>;
 	close($fh);
         $errstr = "";
@@ -666,7 +680,7 @@ sub users {
 sub users_from_shadow {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my @a;
-	open(my $fh, '<', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $!"; return };
+	open(my $fh, '<', $self->shadow_file()) or do { $errstr = "Can't open shadow file ".$self->shadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	push @a, (split(/:/,$_))[0] while <$fh>;
 	close($fh);
 	$errstr = "";
@@ -690,8 +704,8 @@ sub del_group {
 
 	my @dels;
 	my $tmp = $self->group_file.'.tmp';
-	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
-	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+	open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 	chmod PERM_GRP, $ch;
 	while(my $line = <$fh>){
 		my ($name) = split(/:/,$line,2);
@@ -699,19 +713,19 @@ sub del_group {
 		else{ print $ch $line; }
 	}
 	close($fh);close($ch);
-	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $!"; return };
+	move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 	if(-f $self->gshadow_file){
 		my $tmp = $self->gshadow_file.'.tmp';
-		open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $!"; return };
-		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
-		chmod PERM_SHD, $ch;
+		open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
+		chmod PERM_GSD, $ch;
 		while(my $line = <$fh>){
 			my ($name) = split(/:/,$line,2);
 			print $ch $line if $group ne $name;
 		}
 		close($fh);close($ch);
-		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $!"; return };
+		move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	}
 
 	umask $umask;
@@ -763,8 +777,8 @@ sub group {
 
 		my $mod;
 		my $tmp = $self->group_file.'.tmp';
-		open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
-		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
+		open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+		open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
 		chmod PERM_GRP, $ch;
 		while(my $line = <$fh>){
 			chomp $line;
@@ -776,14 +790,14 @@ sub group {
 		}
 		print $ch join(q/:/, $group, 'x', $gid, join(q/,/, @$users)),"\n" unless $mod;
 		close($fh);close($ch);
-		move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $!"; return };
+		move($tmp, $self->group_file()) or do { $errstr = "Can't replace group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 
 		if(-f $self->gshadow_file){
 			my $mod;
 			my $tmp = $self->gshadow_file.'.tmp';
-			open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $!"; return };
-			open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $!"; return };
-			chmod PERM_SHD, $ch;
+			open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
+			open(my $ch, '>', $tmp) or do { $errstr = "Can't open tmp file $tmp: $! (".__FILE__." line ".__LINE__.")"; return };
+			chmod PERM_GSD, $ch;
 			while(my $line = <$fh>){
 				chomp $line;
 				my ($name, $passwd) = split(/:/,$line,3);
@@ -794,13 +808,13 @@ sub group {
 			}
 			print $ch join(q/:/, $group, '!', q//, join(q/,/, @$users)),"\n" unless $mod;
 			close($fh);close($ch);
-			move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $!"; return };
+			move($tmp, $self->gshadow_file()) or do { $errstr = "Can't replace gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 		}
 
 		umask $umask;
 	}else{
 		my ($gid, @users);
-		open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
+		open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 		while(my $line = <$fh>){
 			chomp $line;
 			my ($name, undef, $id, $usrs) = split(/:/,$line,4);
@@ -817,7 +831,7 @@ sub group {
                     return undef, [ ];
                 }
 
-		open($fh, '<', $self->passwd_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
+		open($fh, '<', $self->passwd_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 		while(my $line = <$fh>){
 			my ($login, undef, undef, $id) = split(/:/,$line,5);
 			next if $id != $gid;
@@ -840,7 +854,7 @@ sub group {
 sub groups {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my @a;
-	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $!"; return };
+	open(my $fh, '<', $self->group_file()) or do { $errstr = "Can't open group file ".$self->group_file.": $! (".__FILE__." line ".__LINE__.")"; return };
         push @a, (split(/:/,$_))[0] while <$fh>;
 	close($fh);
 	return @a;
@@ -849,7 +863,7 @@ sub groups {
 sub groups_from_gshadow {
 	my $self = scalar @_ && ref $_[0] eq __PACKAGE__ ? shift : $Self;
 	my @a;
-	open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $!"; return };
+	open(my $fh, '<', $self->gshadow_file()) or do { $errstr = "Can't open gshadow file ".$self->gshadow_file.": $! (".__FILE__." line ".__LINE__.")"; return };
 	push @a, (split(/:/,$_))[0] while <$fh>;
 	close($fh);
 	$errstr = "";
@@ -868,7 +882,7 @@ Passwd::Unix::Alt - Manipulate /etc/{passwd,shadow,group,gshadow} entries
 
 =head1 VERSION
 
-version 0.5203
+version 0.5204
 
 =head1 SYNOPSIS
 
